#include <cmath>
#include "RenderChimp.h"
#include "SceneGraph.h"
#include "Transformable.h"
#include "Camera.h"

/* Warnings */
static const Result_t WRN_CAM_ARG = {
	RESULT_WARNING, "Camera values are not valid" };

static const Result_t WRN_CAM_FOV = {
	RESULT_WARNING, "Very large FOV. The FOV value should be given in radians [0, PI], and not in degrees." };

/*---------------------------------------------------------------------------*/

Camera::Camera(
		char	*nname
	) : Transformable(nname, NODE_CAMERA)
{
	projectionMatrix = mat4f(1, 0, 0, 0,
							 0, 1, 0, 0,
							 0, 0, 1, 0,
							 0, 0, 0, 1);
	nearPlane = 1;
	farPlane = 100;
	projectionType = PROJECTION_TYPE_FREE_TRANSFORM;
}

/*---------------------------------------------------------------------------*/

Camera::~Camera()
{

}

/*---------------------------------------------------------------------------*/

void Camera::setOrthogonalProjection(
		f32		left,
		f32		right,
		f32		bottom,
		f32		top,
		f32		nnear,
		f32		nfar
	)
{
	f32 arl, srl_inv;
	f32 atb, stb_inv;
	f32 afn, sfn_inv;

	if (left == right || top == bottom || nfar == nnear) {
		REPORT(WRN_CAM_ARG);
		return;
	}

	arl = right + left;
	srl_inv = 1.0f / (right - left);
	atb = top + bottom;
	stb_inv = 1.0f / (top - bottom);
	afn = nfar + nnear;
	sfn_inv = 1.0f / (nfar - nnear);

	projectionMatrix.identity();

	projectionMatrix.R00 = 2.0f * srl_inv;
	projectionMatrix.R11 = 2.0f * stb_inv;
	projectionMatrix.R22 = - 2.0f * sfn_inv;

	projectionMatrix.T0 = - arl * srl_inv;
	projectionMatrix.T1 = - atb * stb_inv;
	projectionMatrix.T2 = - afn * sfn_inv;

	projectionType = PROJECTION_TYPE_ORTHOGONAL;
}

/*---------------------------------------------------------------------------*/

void Camera::setPerspectiveProjection(
		f32		fovy,
		f32		aspect,
		f32		nnear,
		f32		nfar
	)
{
	f32 f, t;

	if (fovy > 3.14f) {
		REPORT(WRN_CAM_FOV);
		return;
	}

	if (nnear >= nfar || nnear < 0.01f || nfar < 0.0f) {
		REPORT(WRN_CAM_ARG);
		return;
	}

	f = 1.0f / tan(fovy * 0.5f);
	t = 1.0f / (nnear - nfar);

	projectionMatrix.identity();

	projectionMatrix.R00 = f / aspect;
	projectionMatrix.R11 = f;
	projectionMatrix.R22 = (nnear + nfar) * t;
	projectionMatrix.T2 = 2.0f * nnear * nfar * t;
	projectionMatrix.N2 = -1.0f;
	projectionMatrix.N3 = 0.0f;

	projectionType = PROJECTION_TYPE_PERSPECTIVE;
}

/*---------------------------------------------------------------------------*/

void Camera::setProjection(
		mat4f	&matrix
	)
{
	projectionMatrix = matrix;
	projectionType = PROJECTION_TYPE_FREE_TRANSFORM;
}


/*---------------------------------------------------------------------------*/

void Camera::calculateViewTransform(mat4f *matrix)
{
	matrix->identity();

	transformCamera(matrix);
}

/*---------------------------------------------------------------------------*/

Node *Camera::duplicateSelf()
{
	Camera *c = sceneGraph.createCamera(name);

	c->nearPlane = nearPlane;
	c->farPlane = farPlane;
	c->projectionMatrix = projectionMatrix;
	c->projectionType = projectionType;

	return c;
}

/*---------------------------------------------------------------------------*/

