#ifndef __RC_NODE_H__
#define __RC_NODE_H__

/* Node types */
typedef enum {

	NODE_CAMERA = 0,
	NODE_GROUP,
	NODE_LIGHT,
	NODE_STATE,
	NODE_WORLD,
	NODE_TRIANGLEMESH,
	NODE_SPRITE,

	NODE_TYPE_LAST

} NodeType_t;


class Node {
	friend class SceneGraph;
	friend class Transformable;
	friend class State;
	friend class Light;
	friend class Geometry;

/*----------------------------------------------------------[ Functions ]----*/

	protected:
		Node(
				const char	*nname,
				NodeType_t	ntype
			);

		/*
		 * Delete this node and its contents.
		 * This also recursively deletes the children of the node.
		 */
		virtual ~Node();

	public:
		/* Change the name of the node */
		void setName(
				char		*nname
			);

		/* Get the number of children */
		u32 getChildrenCount();

		/*
		 * Attach a child to this node. The child musn't already have parent.
		 * If it has, detach it first.
		 */
		void attachChild(
				Node		*node
			);

		/* Detaches the child node from this node. */
		void detachChild(
				Node		*node
			);

		/* Detach this node from its parent */
		void detachFromParent();

		/*
		 * Get the child on position "index". Returns 0 if no child is
		 * found.
		 */
		Node *getChild(
				u32			index
			);

		/*
		 * Search for a child with a certain name. Returns 0 if the child
		 * is not found.
		 */
		Node *getChild(
				char		*search_name,
				u32			max_depth
			);

		/* Get the next node of the same parent */
		Node *getNextSibling();

		/* Return parent node */
		Node *getParent();

		/*
		 * Mark this node as active or inactive.
		 * When drawing the scene graph, inactive nodes terminate the
		 * recursive traversal, so none of its children will be drawn either.
		 */
		void setActive(
				bool		active
			);

		/*
		 * Mark this node as visible or invisible.
		 * An invisible node is not drawn during a draw()-call, but the node's
		 * transforms are calculated and its children are drawn.
		 */
		void setVisible(
				bool		visible
			);

		/* Get the name of the node */
		char *getName();

		/*
		 * Duplicates the node, which means that a new node is created, but all
		 * the resources are instantiated, and not cloned.
		 * Only the base node changes name to new_name, all its children will
		 * still have their old name.
		 */
		Node *duplicate(
				char		*new_name,
				bool		recursively		/* True = recursively duplicate all
											 * children as well
											 */
			);

	protected:

		/* Recursive transform camera function */
		virtual void transformCamera(
				mat4f		*matrix
			);

		/* Update matrices */
		virtual void update(
				FrameInfo_t			*di
			);

		/* Recursively draw self and all children */
		virtual void draw(
				FrameInfo_t			*di
			);

		/* Draw the current node */
		virtual void drawSelf(
				FrameInfo_t			*di
			) { }

		virtual Node *duplicateSelf();

		/* Recursively find nodes that references "light" and dereference it */
		/* Semi-hack */
		virtual void dereferenceLight(
				Light				*light
			);


/*------------------------------------------------------------[ Members ]----*/

	protected:
		char		*name;
		NodeType_t	type;

		Node		*parent;
		Node		*firstChild;
		Node		*next;
		u16			nChildren;

	public:
		bool		isActive;		/* False = not traversed */
		bool		isVisible;		/* False = traversed but not drawn */

};




#endif /* __RC_NODE_H__ */
