/*===========================================================================*/

/*
 *           _
 *          c "    -[ RenderChimp ]-
 *           (-      -[ v  0.2a ]-
 *
 *
 *   Written by Magnus Andersson (magnusa@cs.lth.se)
 *   With the help of Rasmus Barringer, Michael Doggett and Bjrn Johnsson
 *
 */

/*===========================================================[ Overview ]====*/

/*-------------------------------------------------------[ Introduction ]----*/

/*
 *   RenderChimp is a (more or less) platform independent framework built to
 * simplify OpenGL ES 2.0 development. The framework basically consists of a
 * scene graph implementation, with supporting systems. RenderChimp is designed
 * in such a manner that the introduction of new features a fairly simple
 * endevour.
 *
 *   A scene graph is a graph representation of all the objects and states in
 * a game or application. The implementation of the scene graph found here is
 * just one solution - it can be extended, optimized or completely rethought.
 * There are many more scene graphs out there, with very different design
 * choices. It's really up to your application which works the best.
 *
 *   For this framework, however, we have strived towards a general and simple
 * solution that can be used for a wide variety of applications.
 *
 */

/*--------------------------------------------------[ Scene Graph Nodes ]----*/

/*
 *   The scene graph in RenderChimp actually has more of a tree like structure
 * than a graph. The main building blocks of the tree are called nodes. Most
 * often these nodes have a clear parent-child type relation.
 *
 *   The node types that can be used to describe a scene are listen below.
 * The nodes share functionality through inheritance. Node surrounded by < >-
 * brackets are abstract, and cannot be created.
 *
 *
 *               <Node>
 *
 *                  |
 *
 *            /     |     \
 *
 *        World   State   <Transformable>
 *
 *                               |
 *
 *                     /       /   \      \
 *
 *                 Group  Camera   Light  <Geometry>
 *
 *                                             |
 *
 *                                          /     \
 *
 *                                     Sprite     TriangleMesh
 *
 *
 *   Node - This is the base class. All nodes inherit from this node type. It
 * mostly contains functionality to handle graph structure, such as attaching
 * and detaching child nodes, etc, which is common to all nodes.
 *
 *   World - The root object of your scene graph. Calling drawAll() on this
 * object will begin the traversal and rendering of the whole scene graph.
 *
 *   State - Use this to change the internal render state (such as blend mode,
 * depth testing, culling, etc).
 *
 *   Transformable - Base class for nodes that can be transformed (rotated,
 * translated and scaled).
 *
 *   Group - Doesn't really do much. Use this to stack transformations.
 *
 *   Camera - A camera with orthogonal, perspective or free projection.
 *
 *   Light - Describes a simple point light with intensity, color and position.
 *
 *   Geometry - Base class for different geometry.
 *
 *   Sprite - A quad (with a texture) that orients itself so that it is always
 * facing the camera. Great for explosions!
 *
 *   TriangleMesh - This node is probably the most potent one. It describes a
 * triangle mesh, which comprises all of what is rendered (Sprites are in fact
 * a special variation of a triangle mesh).
 *
 */

/*----------------------------------------------[ Scene Graph Resources ]----*/

/*
 *   Nodes alone are not enough to describe a scene. We havn't yet described
 * how textures, vertex arrays and shaders, for example, are handled. For this
 * we introduce a new concept - resources.
 *
 *   Unlike nodes, resources do not follow a tree-like structure. It is very
 * common that the same texture is used over and over again, or that there are
 * many instances of the same triangle mesh. A texture can consist of many
 * pixels and a mesh can have large vertex and index arrays. It would be very
 * wasteful to have several copies of the same texture or mesh, that share
 * the same data.
 *
 *   A resource is a container for a chunk of data that is used by multiple
 * objects. Instead of creating many copies, the resource list system keeps
 * track of all the resources and only allow one copy of each unique data set
 * (unless the data is explicitly cloned).
 *
 *   The nodes in the scene graph are free to reference these resources and use
 * them for their various purposes. The different resources that are available
 * are:
 *
 *   Vertex array - A vertex array is a list of per-vertex attributes such as
 * position, normal, color, texture coordinate, etc. These are organized in an
 * "array of structure" manner, which means that all of the attributes for one
 * vertex are grouped together in the list (as opposed to having one whole
 * array for each attribute type (= bad cache perfomance)).
 *
 *   Index array - We don't want to duplicate all of the vertex attributes for
 * each of the triangles in a mesh. Instead an index array is used that points
 * to the indices in the vertex array to use for each triangle. Currently, all
 * triangles are comprised of three indices (i.e. no triangle strips), so there
 * are three times as many indices in this array as there are triangles in the
 * mesh.
 *
 *   Material - Materials define the appearance of a mesh, be it texturing,
 * BRDF, displacement, bumpyness, or whatever neat thing you can think of. All
 * materials is associated with the same ber shader program, with different
 * code paths turned on or off. This shader utilizes the GPU to achieve the
 * desired appearance.
 *
 *   Shader - A shader is either a vertex or a fragment program that is used by
 * one or many shader programs.
 *
 *   ShaderProgram - A combination of a vertex and a fragment shader comprises
 * a shader program. A shader program is run on the GPU and it's up to the
 * application to feed it with the correct data.
 *
 *   Texture - Textures can be used for a variety of purposes. The most common
 * usage is as a diffuse component in the BRDF calculation of a material,
 * but it isn't, by any means, limited to that.
 *
 */

/*--------------------------------------------------[ Other sub systems ]----*/

/*
 *   To complement the scene graph functionality, a few other components have
 * been implemented in to the RenderChimp framework.
 *
 *   Vector and matrix math - Most of the basic vector and matrix operations
 * have been implemented and incorporated in a set of new data types.   
 *
 *   Memory debugger - A memory debugger help keeping track of memory usage,
 * heap corruption and memory leaks. It's highly recommended that this is
 * switched on from time to time, just to make sure that no new memory
 * abnormalities are present. Keep in mind, though, that the memory debugger
 * quite efficiently drains the performance of your application.
 *
 *   Text - A (quick and dirty) text printer is available, which simply prints
 * bitmap font text on screen. Useful for debugging when concole output is
 * cumbersome or unavailable.
 *
 *   Platform backend - Platform specific function calls and device simulation
 * is moved down to a layer that is "below" the scene graph. This includes
 * system time fetching, input, texture reading, etc.
 *
 */

/*---------------------------------------------------------------[ .pwn ]----*/

/*
 *   It should also be mentioned that RenderChimp supports its own scene graph
 * format called ".pwn". The framework doesn't contain any converter or writer
 * for the format, only a parser. There is a Collada-to-pwn converter available
 * as a separate project, which works fairly well for simple scenes.
 *
 */

/*----------------------------------------------------[ Getting started ]----*/

/*
 *   Your best bet of getting in to this framework quickly is to open up the
 * file SceneGraph.h and explore what different nodes and resources are
 * available. You can then go from there and examine each of these in more
 * detail in their respective .h file.
 *
 *   Good luck!
 */

/*=====================================================[ Build settings ]====*/

#ifndef __RC_RENDERCHIMP_H__
#define __RC_RENDERCHIMP_H__

/* Win32 build flag (WIN32)
 * 
 * Enable to build for 32-bit Windows.
 * Either build for Mac or for Win, not both.
 */
/* Define WIN32 in project settings */

/* Mac build flag (MAC)
 * 
 * Enable to build for 32-bit MacOS X.
 * Either build for Mac or for Win, not both.
 */
/* Define MAC in project settings */

/* Memory debug
 * 
 * Enable definition to debug memory.
 * This will enable tracking and validation of all memory allocated on the
 * heap (at the expense of performance).
 */
#define MEMORY_DEBUG

/* Log to file
 *
 * If this definition is enabled warnings, errors and memory debug info will
 * be written to log.txt and mem.txt respectively. If this is disabled only
 * standard output will be used to log this information.
 */
#define LOG_TO_FILE

/* Get rid of dirty Mirco$oft warnings */
#ifdef WIN32
	#pragma warning( disable : 4996 )
	#pragma warning( disable : 4201 )
	#pragma warning( disable : 4100 )
#endif

/*=========================================================[ Tweakables ]====*/

#define MAX_ATTRIBS								16
#define MAX_SHADER_LIGHT_SLOTS					4

#define MATERIAL_MAX_ALLOWED_TEXTURES			8

/*=================================================[ Essential includes ]====*/

#ifdef WIN32
	#include <EGL/egl.h>
	#include <GLES2/gl2.h>
#elif defined MAC

#endif /* WIN32 / MAC */

#include <vector>
#include <memory>
#include <memory>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#include "Types.h"

#include "Memory.h"

#include "VectorMath.h"
#include "Result.h"
#include "Misc.h"

#include "SceneGraph.h"

/*===========================================================================*/

#endif /* __RC_RENDERCHIMP_H__ */

/*===========================================================================*/

