/*
 * Defines an internal State.
 *
 * This sets the rendering state. The state specified in a State node will
 * hold for the whole subtree of the node, or until a new State node is comes
 * along.
 */

#ifndef __RC_RENDERSTATE_H__
#define __RC_RENDERSTATE_H__

/*-----------------------------------------------------[ State settings ]----*/

/* Comparation strategy for depth and stencil */
typedef enum {

	CMP_NEVER = 0,
	CMP_LESS,
	CMP_EQUAL,
	CMP_LEQUAL,
	CMP_GREATER,
	CMP_NOTEQUAL,
	CMP_GEQUAL,
	CMP_ALWAYS,

	CMP_LAST

} StateCompare_t;


/* Source blending methods */
typedef enum {

	SB_ZERO = 0,
	SB_ONE,
	SB_SRC_COLOR,
	SB_ONE_MINUS_SRC_COLOR,
	SB_SRC_ALPHA,
	SB_ONE_MINUS_SRC_ALPHA,
	SB_DST_ALPHA,
	SB_ONE_MINUS_DST_ALPHA,

	SB_LAST

} StateSrcBlend_t;

/* Destination blending methods */
typedef enum {

	DB_ZERO = 0,
	DB_ONE,
	DB_DST_COLOR,
	DB_ONE_MINUS_DST_COLOR,
	DB_SRC_ALPHA_SATURATE,
	DB_ONE_MINUS_SRC_ALPHA,
	DB_DST_ALPHA,
	DB_ONE_MINUS_DST_ALPHA,

	DB_LAST

} StateDstBlend_t;

/* Culling modes */
typedef enum {

	CULL_FRONT = 0,
	CULL_BACK,
	CULL_FRON_AND_BACK,

	CULL_LAST

} StateCulling_t;

/*---------------------------------------------------------------------------*/

/* Render state description */
typedef struct {

	i32					stencilRef;
	u32					stencilMask;
	StateCompare_t		stencilTest;

	StateCompare_t		depthTest;

	StateSrcBlend_t		srcBlend;
	StateDstBlend_t		dstBlend;

	StateCulling_t		cull;

	i32					scissorX, scissorY;
	u32					scissorW, scissorH;

	u8					changes;
	u8					flags;

} StateSettings_t;



/*---------------------------------------------------------------------------*/


class RenderState {
	friend class SceneGraph;
	friend class World;
	friend class State;

/*----------------------------------------------------------[ Functions ]----*/

	protected:
		RenderState();

		~RenderState();

	public:

		/* Enable/disable depth testing */

		void enableDepth(
				StateCompare_t	depth_test
			);

		void disableDepth();


		/* Enable/disable stencil testing */

		void enableStencil(
				StateCompare_t	stencil_test,
				i32				ref,
				u32				mask
			);

		void disableStencil();


		/* Enable/disable blending */

		void enableBlend(
				StateSrcBlend_t	src_blend,
				StateDstBlend_t	dst_blend
			);

		void disableBlend();


		/* Enable/disable triangle culling */

		void enableCulling(
				StateCulling_t	cull_strategy
			);

		void disableCulling();


		/* Enable/disable scissor test */

		void enableScissor(
				i32				x,
				i32				y,
				u32				width,
				u32				height
			);

		void disableScissor();

		void clear();

		static void setState(
				StateSettings_t			*curr,
				u8						changes,
				bool					force_set
			);

		static StateSettings_t *getCurrentState();

/*------------------------------------------------------------[ Members ]----*/

	private:
		StateSettings_t			settings;

};


#endif /* __RC_RENDERSTATE_H__ */

