#include "RenderChimp.h"
#include "SceneGraph.h"
#include "Resource.h"
#include "ResourceList.h"
#include "ShaderProgram.h"

#define SHADER_PROGRAM_ERROR_BUFFER_LENGTH			16384

static u16 shader_clone_counter		= 0;

/* Warnings */
static const Result_t WRN_LINK = {
	RESULT_WARNING, "Couldn't link shaders" };

static const Result_t WRN_CLONE = {
	RESULT_WARNING, "Cloning a shader program resource is not permitted" };

/*---------------------------------------------------------------------------*/

ShaderProgram::ShaderProgram(
		const char			*nidentifier,
		Shader				*vertex_shader,
		Shader				*fragment_shader,
		u32					purge_level
	) : Resource(nidentifier, purge_level, RESOURCE_SHADER_PROGRAM)
{
	GLint status;

	if (strlen(nidentifier) == 0)
		REPORT(ERR_ARG);

	GL_ASSERT();

	vertexShader = vertex_shader;
	fragmentShader = fragment_shader;

	id = glCreateProgram();

	GL_ASSERT();

	glAttachShader(id, vertexShader->id);
	glAttachShader(id, fragmentShader->id);
	
	GL_ASSERT();

	glLinkProgram(id);

	GL_ASSERT();

	glGetProgramiv(id, GL_LINK_STATUS, &status);

	GL_ASSERT();

	if (status == GL_FALSE) {
		GLsizei length;
		char result[SHADER_PROGRAM_ERROR_BUFFER_LENGTH];

		REPORT(WRN_LINK);

		glGetProgramInfoLog(id, SHADER_PROGRAM_ERROR_BUFFER_LENGTH, &length, result);

		GL_ASSERT();

		if (length) {
			Result_t msg = {RESULT_ERROR, result};

			REPORT(msg);
		}
	}

	GL_ASSERT();

	size = vertexShader->size + fragmentShader->size;
}

/*---------------------------------------------------------------------------*/

ShaderProgram::~ShaderProgram()
{
	if (vertexShader)
		sceneGraph.dereferenceResource(vertexShader);

	if (fragmentShader)
		sceneGraph.dereferenceResource(fragmentShader);
}

/*---------------------------------------------------------------------------*/

void ShaderProgram::activate()
{
	glUseProgram(id);

	GL_ASSERT();
}

/*---------------------------------------------------------------------------*/

GLint ShaderProgram::getAttributeLocation(
		const char			*name
	)
{
	return glGetAttribLocation(id, name);
}

/*---------------------------------------------------------------------------*/

bool ShaderProgram::setAttribute(
		const char			*name,
		const f32			*values,
		u32					dimension
	)
{
	GLint loc = getAttributeLocation(name);

	if (loc == -1)
		return false;

	switch (dimension) {
		case 1:
			glVertexAttrib1fv(loc, values);
			break;
		case 2:
			glVertexAttrib2fv(loc, values);
			break;
		case 3:
			glVertexAttrib3fv(loc, values);
			break;
		case 4:
			glVertexAttrib4fv(loc, values);
			break;
		default:
			return false;
			break;
	}

	return true;
}

/*---------------------------------------------------------------------------*/

bool ShaderProgram::setAttributePointer(
		const char			*name,
		u32					size,
		bool				normalized,
		u32					stride,
		const void			*ptr
	)
{
	i32 loc = glGetAttribLocation(id, name);

	if (loc == -1)
		return false;

	glVertexAttribPointer(loc, size, GL_FLOAT, normalized ? GL_TRUE : GL_FALSE, stride * sizeof(f32), ptr);

	glEnableVertexAttribArray(loc);

	GL_ASSERT();

	return true;
}

/*---------------------------------------------------------------------------*/

GLint ShaderProgram::getUniformLocation(
		const char			*name
	)
{
	return glGetUniformLocation(id, name);
}

/*---------------------------------------------------------------------------*/

bool ShaderProgram::setUniform(
		const char			*name,
		const i32			*values,
		u32					dimension
	)
{
	GLint loc = getUniformLocation(name);

	GL_ASSERT();

	if (loc == -1)
		return false;

	switch (dimension) {
		case 1:
			glUniform1i(loc, values[0]);
			break;
		case 2:
			glUniform2i(loc, values[0], values[1]);
			break;
		case 3:
			glUniform3i(loc, values[0], values[1], values[2]);
			break;
		case 4:
			glUniform4i(loc, values[0], values[1], values[2], values[3]);
			break;
		default:
			return false;
			break;
	}

	GL_ASSERT();

	return true;
}

/*---------------------------------------------------------------------------*/

bool ShaderProgram::setUniform(
		const char			*name,
		const f32			*values,
		u32					dimension
	)
{
	GLint loc = getUniformLocation(name);

	GL_ASSERT();

	if (loc == -1)
		return false;

	switch (dimension) {
		case 1:
			glUniform1f(loc, values[0]);
			break;
		case 2:
			glUniform2f(loc, values[0], values[1]);
			break;
		case 3:
			glUniform3f(loc, values[0], values[1], values[2]);
			break;
		case 4:
			glUniform4f(loc, values[0], values[1], values[2], values[3]);
			break;
		default:
			return false;
			break;
	}

	GL_ASSERT();

	return true;
}

/*---------------------------------------------------------------------------*/

bool ShaderProgram::setUniformVector(
		const char			*name,
		const i32			*values,
		u32					count,
		u32					dimension
	)
{
	GLint loc = getUniformLocation(name);

	GL_ASSERT();

	if (loc == -1)
		return false;

	switch (dimension) {
		case 1:
			glUniform1iv(loc, count, values);
			break;
		case 2:
			glUniform2iv(loc, count, values);
			break;
		case 3:
			glUniform3iv(loc, count, values);
			break;
		case 4:
			glUniform4iv(loc, count, values);
			break;
		default:
			return false;
			break;
	}

	GL_ASSERT();

	return true;
}

/*---------------------------------------------------------------------------*/

bool ShaderProgram::setUniformVector(
		const char			*name,
		const f32			*values,
		u32					count,
		u32					dimension
	)
{
	GLint loc = getUniformLocation(name);

	GL_ASSERT();

	if (loc == -1)
		return false;

	switch (dimension) {
		case 1:
			glUniform1fv(loc, count, values);
			break;
		case 2:
			glUniform2fv(loc, count, values);
			break;
		case 3:
			glUniform3fv(loc, count, values);
			break;
		case 4:
			glUniform4fv(loc, count, values);
			break;
		default:
			return false;
			break;
	}

	GL_ASSERT();

	return true;
}

/*---------------------------------------------------------------------------*/

bool ShaderProgram::setUniformMatrix(
		const char			*name,
		const f32			*values,
		u32					count,
		u32					dimension
	)
{
	GLint loc = getUniformLocation(name);

	GL_ASSERT();

	if (loc == -1)
		return false;

	switch (dimension) {
		case 2:
			glUniformMatrix2fv(loc, count, false, values);
			break;
		case 3:
			glUniformMatrix3fv(loc, count, false, values);
			break;
		case 4:
			glUniformMatrix4fv(loc, count, false, values);
			break;
		default:
			return false;
			break;
	}

	GL_ASSERT();

	return true;
}

/*---------------------------------------------------------------------------*/

bool ShaderProgram::setTextureSampler(
		const char			*name,
		Texture				*tex,
		u32					slot
	)
{
	GLint loc;

	glActiveTexture(GL_TEXTURE0 + slot);
	glBindTexture(GL_TEXTURE_2D, tex->getGLTexId());

	GL_ASSERT();

	loc = getUniformLocation(name);

	GL_ASSERT();

	if (loc == -1)
		return false;

	glUniform1i(loc, slot);

	GL_ASSERT();

	return true;
}

/*---------------------------------------------------------------------------*/

bool ShaderProgram::setCubeMapSampler(
		const char			*name,
		CubeMap				*cubeMap,
		u32					slot
	)
{
	GLint loc;

	glActiveTexture(GL_TEXTURE0 + slot);
	glBindTexture(GL_TEXTURE_CUBE_MAP, cubeMap->getGLTexId());

	GL_ASSERT();

	loc = getUniformLocation(name);

	GL_ASSERT();

	if (loc == -1)
		return false;

	glUniform1i(loc, slot);

	GL_ASSERT();

	return true;
}

/*---------------------------------------------------------------------------*/

Resource *ShaderProgram::clone(
		char			*clone_name,

		u32				purge_level
	)
{
	REPORT(WRN_CLONE);

	return 0;
}

/*---------------------------------------------------------------------------*/

GLuint ShaderProgram::getID()
{
	return id;
}

/*---------------------------------------------------------------------------*/

#ifdef DEBUG

void ShaderProgram::print(
		FILE		*f
	)
{
	char *c = vertexShader->getIdentityStringCopy();
	char *d = fragmentShader->getIdentityStringCopy();

#ifdef LOG_TO_FILE
	#define prnt0(a)		fprintf(f, a)
	#define prnt1(a, b)		fprintf(f, a, b)
#else /* LOG_TO_FILE */
	#define prnt0(a)		printf(a)
	#define prnt1(a, b)		printf(a, b)
#endif /* LOG_TO_FILE */

	prnt1("    [ V.SHADER ][ %s ]\n", c);
	prnt1("    [ F.SHADER ][ %s ]\n", d);
	prnt1("    [ GL ID    ][ %u ]\n", id);

	free(c);
	free(d);
}

#endif /* DEBUG */

/*---------------------------------------------------------------------------*/

