#include "RenderChimp.h"
#include "SceneGraph.h"
#include "Geometry.h"
#include "Sprite.h"

/*---------------------------------------------------------------------------*/

void Sprite::setup_sprite(
		const char				*nname,
		Texture					*ntexture,

		bool					instantiate_resources
	)
{
	type = NODE_SPRITE;

	vertexArray = sceneGraph.getVertexArray("SpriteVertexArray");

	if (!vertexArray) {
		float v[4 * 2];

		v[0] = -1.0f;
		v[1] = -1.0f;

		v[2] = 1.0f;
		v[3] = -1.0f;

		v[4] = -1.0f;
		v[5] = 1.0f;

		v[6] = 1.0f;
		v[7] = 1.0f;

		vertexArray = sceneGraph.createVertexArray("SpriteVertexArray", v, 2, 4, NO_PURGE);
		vertexArray->setAttribute(0, 2, "Vertex");
	}

	indexArray = sceneGraph.getIndexArray("SpriteIndexArray");

	if (!indexArray) {
		u16	i[6];

		i[0] = 0;
		i[1] = 1;
		i[2] = 2;
		i[3] = 1;
		i[4] = 3;
		i[5] = 2;

		indexArray = sceneGraph.createIndexArray("SpriteIndexArray", i, 6, TRIANGLES, NO_PURGE);
	}

	size.x = 1.0f;
	size.y = 1.0f;

	texcoordsFrom = vec2f(0.0f, 0.0f);
	texcoordsTo = vec2f(1.0f, 1.0f);

	shaderProgram = sceneGraph.createShaderProgram("SpriteProg", "\n", "../shaders/Sprite.vs", "../shaders/Sprite.fs", true);

	if (instantiate_resources) {
		texture = (Texture *) ntexture->instantiate();
	} else {
		texture = ntexture;
	}
}

/*---------------------------------------------------------------------------*/

Sprite::Sprite(
		const char				*nname,
		Texture					*ntexture,

		bool					instantiate_resources
	) : Geometry(nname, NODE_SPRITE)
{
	setup_sprite(nname, ntexture, instantiate_resources);
}

/*---------------------------------------------------------------------------*/

Sprite::Sprite(
		const char				*nname,
		const char				*texture_filename,

		u32						purge_level
	) : Geometry(nname, NODE_SPRITE)
{
	Texture *ntexture = sceneGraph.createTexture(texture_filename,
		true, TEXTURE_FILTER_TRILINEAR, TEXTURE_WRAP_CLAMP, purge_level);

	setup_sprite(nname, ntexture, false);
}

/*---------------------------------------------------------------------------*/

Sprite::~Sprite()
{
	sceneGraph.dereferenceResource(vertexArray);
	sceneGraph.dereferenceResource(indexArray);
}

/*---------------------------------------------------------------------------*/

void Sprite::drawSelf(
		FrameInfo_t				*di
	)
{
	u32 i;
	vec3f v;

	rgbf light_color = vec3f(1.0f, 1.0f, 1.0f);

	GL_ASSERT();

	if (di->activeShader != shaderProgram) {
		shaderProgram->activate();
		di->activeShader = shaderProgram;
	}

	di->matTransform->setTransformationMatrices(shaderProgram);

	v.x = world.T0;
	v.y = world.T1;
	v.z = world.T2;

	shaderProgram->setUniformVector("Size", size.s, 1, 2);
	shaderProgram->setUniformVector("Position", v.s, 1, 3);
	shaderProgram->setUniformVector("TexcoordsFrom", texcoordsFrom.s, 1, 2);
	shaderProgram->setUniformVector("TexcoordsTo", texcoordsTo.s, 1, 2);

	shaderProgram->setTextureSampler("TextureSampler", texture, 0);

	for (i = 0; i < MAX_SHADER_LIGHT_SLOTS; i++) {
		if (light[i]) {
			f32 intensity = light[i]->intensity;

			light_color = light[i]->color * intensity;
			
			break;
		}
	}

	shaderProgram->setUniformVector("LightColor", light_color.s, 1, 3);

	GL_ASSERT();

	vertexArray->pushToShader(shaderProgram);

	glDrawElements(GL_TRIANGLES, indexArray->getLength(), GL_UNSIGNED_SHORT, indexArray->getArray());

	GL_ASSERT();
}

/*---------------------------------------------------------------------------*/

void Sprite::setTexture(
		Texture					*tex,

		bool					instantiate_resource
	)
{
	Texture::setTexture(&texture, tex, instantiate_resource);
}

/*---------------------------------------------------------------------------*/

void Sprite::setTexture(
		const char				*filename,

		u32						purge_level
	)
{
	Texture::setTexture(&texture, filename, purge_level);
}

/*---------------------------------------------------------------------------*/

void Sprite::setSize(
		vec2f					nsize
	)
{
	size = nsize;
}

/*---------------------------------------------------------------------------*/

void Sprite::setTexcoords(
		vec2f					from_uv,
		vec2f					to_uv
	)
{
	texcoordsFrom = from_uv;
	texcoordsTo = to_uv;
}

/*---------------------------------------------------------------------------*/

Node *Sprite::duplicateSelf()
{
	u32 i;

	Sprite *c = sceneGraph.createSprite(name, texture, true);

	c->setSize(size);
	c->setTexcoords(texcoordsFrom, texcoordsTo);

	for (i = 0; i < MAX_SHADER_LIGHT_SLOTS; i++) {
		if (light[i] != 0) {
			c->setLight(light[i], i);
		}
	}

	return c;
}

/*---------------------------------------------------------------------------*/

