#include "RenderChimp.h"
#include "SceneGraph.h"
#include "Resource.h"
#include "ResourceList.h"
#include "Texture.h"

/* Warnings */
static const Result_t WRN_CLONE = {
	RESULT_WARNING, "Cloning a texture resource is not permitted" };

static const Result_t WRN_BPP = {
	RESULT_WARNING, "The texture doesn't have 1, 3 or 4 bytes per pixel" };

static const Result_t WRN_OPEN = {
	RESULT_WARNING, "Couldn't open texture file" };

/*---------------------------------------------------------------------------*/

Texture::Texture(
		const char			*filename,
		bool				generate_mipmaps,
		TextureFilter_t		filter,
		TextureWrap_t		wrap,
		u32					purge_level

	) : Resource(filename, purge_level, RESOURCE_TEXTURE)
{
	u8 *pixels;
	TextureFormat_t format;
	u32 w, h, bpp;

	apeGetTexture(filename, &pixels, &w, &h, &bpp);

	if (!pixels) {
		REPORT(WRN_OPEN);
		return;
	}

	if (bpp == 4) {
		format = TEXTURE_FORMAT_RGBA;
	} else if (bpp == 3) {
		format = TEXTURE_FORMAT_RGB;
	} else if (bpp == 1) {
		format = TEXTURE_FORMAT_GRAY;
	} else {
		format = TEXTURE_FORMAT_GRAY; /* Just to get rid of a warning... */
		REPORT(WRN_BPP);
		return;
	}

	if (strlen(filename) == 0)
		REPORT(ERR_ARG);

	setup_texture(pixels, format, w, h, generate_mipmaps, filter, wrap);

	GL_ASSERT();

	apeDestroyTexture(pixels);
}

/*---------------------------------------------------------------------------*/

Texture::Texture(
		const u8			*pixels,
		const char			*nidentifier,
		TextureFormat_t		nformat,
		u32					nwidth,
		u32					nheight,
		bool				generate_mipmaps,
		TextureFilter_t		filter,
		TextureWrap_t		wrap,
		u32					purge_level
	) : Resource(nidentifier, purge_level, RESOURCE_TEXTURE)
{

	if (strlen(nidentifier) == 0)
		REPORT(ERR_ARG);

	setup_texture(pixels, nformat, nwidth, nheight, generate_mipmaps, filter, wrap);
}

/*---------------------------------------------------------------------------*/

void Texture::setup_texture(
		const u8			*pixels,
		TextureFormat_t		nformat,
		u32					nwidth,
		u32					nheight,
		bool				generate_mipmaps,
		TextureFilter_t		filter,
		TextureWrap_t		wrap
	)
{
	width = nwidth;
	height = nheight;

	glGenTextures(1, &texId);
	GL_ASSERT();

	glBindTexture(GL_TEXTURE_2D, texId);
	GL_ASSERT();

	switch (nformat) {
		default:
		case TEXTURE_FORMAT_RGBA:
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, nwidth, nheight, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
			size = nwidth * nheight * 4;
			break;
		case TEXTURE_FORMAT_RGB:
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, nwidth, nheight, 0, GL_RGB, GL_UNSIGNED_BYTE, pixels);
			size = nwidth * nheight * 3;
			break;
		case TEXTURE_FORMAT_GRAY:
			glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE, nwidth, nheight, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, pixels);
			size = nwidth * nheight;
			break;
	}
	GL_ASSERT();

	switch (filter) {
		case TEXTURE_FILTER_NEAREST:
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			break;
		default:
		case TEXTURE_FILTER_BILINEAR:
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			break;
		case TEXTURE_FILTER_TRILINEAR:
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
			break;
	}
	GL_ASSERT();

	switch (wrap) {
		case TEXTURE_WRAP_CLAMP:
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
			break;
		default:
		case TEXTURE_WRAP_REPEAT:
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			break;
		case TEXTURE_WRAP_MIRRORED_REPEAT:
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);
			break;
	}
	GL_ASSERT();

	if (generate_mipmaps)
		glGenerateMipmap(GL_TEXTURE_2D);

	GL_ASSERT();
}

/*---------------------------------------------------------------------------*/

Texture::~Texture()
{
	glDeleteTextures(1, &texId);

	GL_ASSERT();
}

/*---------------------------------------------------------------------------*/

Resource *Texture::clone(
		const char		*clone_name,

		u32				purge_level
	)
{
	REPORT(WRN_CLONE);

	return 0;
}

/*---------------------------------------------------------------------------*/

void Texture::setTexture(
		Texture				**target_tex,
		Texture				*new_tex,
		bool				instantiate_resource
	)
{
	if (*target_tex)
		sceneGraph.dereferenceResource(*target_tex);

	if (instantiate_resource) {
		*target_tex = (Texture *) new_tex->instantiate();
	} else {
		*target_tex = new_tex;
	}
}

/*---------------------------------------------------------------------------*/

void Texture::setTexture(
		Texture				**target_tex,
		const char			*filename,
		u32					purge_level
	)
{
	if (*target_tex)
		sceneGraph.dereferenceResource(*target_tex);

	*target_tex = sceneGraph.createTexture(filename,
		true, TEXTURE_FILTER_TRILINEAR, TEXTURE_WRAP_REPEAT, purge_level);

}

/*---------------------------------------------------------------------------*/

#ifdef DEBUG

void Texture::print(
		FILE		*f
	)
{
#ifdef LOG_TO_FILE
	#define prnt0(a)		fprintf(f, a)
	#define prnt1(a, b)		fprintf(f, a, b)
#else /* LOG_TO_FILE */
	#define prnt0(a)		printf(a)
	#define prnt1(a, b)		printf(a, b)
#endif /* LOG_TO_FILE */

	prnt1("    [ FORMAT   ][ %u ]\n", (u32) format);
	prnt1("    [ WIDTH    ][ %u ]\n", width);
	prnt1("    [ HEIGHT   ][ %u ]\n", height);
	prnt1("    [ MIP'D    ][ %u ]\n", hasMipmaps ? 1 : 0);
	prnt1("    [ GL ID    ][ %u ]\n", texId);
}

#endif /* DEBUG */

/*---------------------------------------------------------------------------*/

