#include "RenderChimp.h"
#include "SceneGraph.h"
#include <cmath>
#include "Transformable.h"

/* Warnings */
static const Result_t WRN_CAMERA = {
	RESULT_WARNING, "Camera resides in a non-active node" };

/*---------------------------------------------------------------------------*/

Transformable::Transformable(
		const char		*nname,
		NodeType_t		ntype
	) : Node(nname, ntype)
{
	reset();
}

/*---------------------------------------------------------------------------*/

Transformable::~Transformable()
{
}

/*---------------------------------------------------------------------------*/

void Transformable::reset()
{
	T.x = 0.0f;
	T.y = 0.0f;
	T.z = 0.0f;

	S.x = 1.0f;
	S.y = 1.0f;
	S.z = 1.0f;

	R.identity();
}

/*---------------------------------------------------------------------------*/

void Transformable::translate(
		f32		x,
		f32		y,
		f32		z
	)
{
	T.x += x;
	T.y += y;
	T.z += z;
}

/*---------------------------------------------------------------------------*/

void Transformable::translate(
		vec3f	&v
	)
{
	translate(v.x, v.y, v.z);
}

/*---------------------------------------------------------------------------*/

void Transformable::scale(
		f32		x,
		f32		y,
		f32		z
	)
{
	S.x *= x;
	S.y *= y;
	S.z *= z;
}

/*---------------------------------------------------------------------------*/

void Transformable::scale(
		vec3f	&v
	)
{
	scale(v.x, v.y, v.z);
}

/*---------------------------------------------------------------------------*/

void Transformable::rotate(
		f32		angle,
		f32		x,
		f32		y,
		f32		z
	)
{
	vec3f v = vec3f(x, y, z);
	mat3f m;

	f32 xx, xy, xz, yy, yz, zz;
	f32 a, c, s;

	if (v.length() != 1.0f) {
		v.normalize();
		x = v.x;
		y = v.y;
		z = v.z;
	}

	a = angle;

	c = cos(a);
	s = sin(a);

	xx = x * x;
	xy = x * y;
	xz = x * z;
	yy = y * y;
	yz = y * z;
	zz = z * z;

	m.a = xx + (1.0f - xx) * c;
	m.b = xy * (1.0f - c) - z * s;
	m.c = xz * (1.0f - c) + y * s;

	m.d = xy * (1.0f - c) + z * s;
	m.e = yy + (1.0f - yy) * c;
	m.f = yz * (1.0f - c) - x * s;

	m.g = xz * (1.0f - c) - y * s;
	m.h = yz * (1.0f - c) + x * s;
	m.i = zz + (1.0f - zz) * c;

	R = m * R;
}

/*---------------------------------------------------------------------------*/

void Transformable::rotate(
		f32		angle,
		vec3f	&v
	)
{
	rotate(angle, v.x, v.y, v.z);
}


/*---------------------------------------------------------------------------*/

void Transformable::setTranslate(
		f32		x,
		f32		y,
		f32		z
	)
{
	T.x = x;
	T.y = y;
	T.z = z;
}

/*---------------------------------------------------------------------------*/

void Transformable::setTranslate(
		vec3f	&v
	)
{
	setTranslate(v.x, v.y, v.z);
}

/*---------------------------------------------------------------------------*/

void Transformable::setScale(
		f32		x,
		f32		y,
		f32		z
	)
{
	S.x = x;
	S.y = y;
	S.z = z;
}

/*---------------------------------------------------------------------------*/

void Transformable::setScale(
		vec3f	&v
	)
{
	setScale(v.x, v.y, v.z);
}

/*---------------------------------------------------------------------------*/

void Transformable::setRotate(
		f32		angle,
		f32		x,
		f32		y,
		f32		z
	)
{
	R.identity();
	rotate(angle, x, y, z);
}

/*---------------------------------------------------------------------------*/

void Transformable::setRotate(
		f32		angle,
		vec3f	&v
	)
{
	setRotate(angle, v.x, v.y, v.z);
}

/*---------------------------------------------------------------------------*/

void Transformable::transformCamera(
		mat4f *matrix
	)
{
	if (!isActive)
		REPORT(WRN_CAMERA);

	/* TODO: Should scale be in here? */
	*matrix = (*matrix) * 
		getScaleMatrixInverse() *
		getRotationMatrixInverse() *
		getTranslationMatrixInverse();

	if (parent)
		parent->transformCamera(matrix);
}

/*---------------------------------------------------------------------------*/

void Transformable::update(
		FrameInfo_t			*di
	)
{
	if (isActive) {
		Node *c = firstChild;
		mat4f *old_world_matrix;

		/* Place old world matrix on the stack */
		old_world_matrix = di->matTransform->world;
		world = *di->matTransform->world * getLocalTransform();
		di->matTransform->world = &world;

		/* Draw children */
		while (c) {
			c->update(di);

			c = c->next;
		}

		/* Reset world matrix from stack */
		di->matTransform->world = old_world_matrix;
	}
}

/*---------------------------------------------------------------------------*/

void Transformable::draw(
		FrameInfo_t			*di
	)
{
	if (isActive) {
		Node *c = firstChild;

		di->matTransform->world = &world;

		if (isVisible) {
			drawSelf(di);
			di->matTransform->finishNode();
		}

		/* Draw children */
		while (c) {
			c->draw(di);

			c = c->next;
		}

	}
}

/*---------------------------------------------------------------------------*/

mat4f Transformable::getTranslationMatrix()
{
	return mat4f(1, 0, 0, T.x,
				 0, 1, 0, T.y,
				 0, 0, 1, T.z,
				 0, 0, 0, 1);
}

/*---------------------------------------------------------------------------*/

mat4f Transformable::getRotationMatrix()
{
	return mat4f(R.a, R.b, R.c, 0,
				 R.d, R.e, R.f, 0,
				 R.g, R.h, R.i, 0,
				 0  , 0  , 0  , 1);
}

/*---------------------------------------------------------------------------*/

mat4f Transformable::getScaleMatrix()
{
	return mat4f(S.x, 0  , 0  , 0,
				 0  , S.y, 0  , 0,
				 0  , 0  , S.z, 0,
				 0  , 0  , 0  , 1);
}

/*---------------------------------------------------------------------------*/

mat4f Transformable::getTranslationMatrixInverse()
{
	return mat4f(1, 0, 0, -T.x,
				 0, 1, 0, -T.y,
				 0, 0, 1, -T.z,
				 0, 0, 0, 1);
}

/*---------------------------------------------------------------------------*/

mat4f Transformable::getRotationMatrixInverse()
{
	return mat4f(R.a, R.d, R.g, 0,
				 R.b, R.e, R.h, 0,
				 R.c, R.f, R.i, 0,
				 0  , 0  , 0  , 1);
}

/*---------------------------------------------------------------------------*/

mat4f Transformable::getScaleMatrixInverse()
{
	return mat4f(1.0f/S.x, 0       , 0       , 0,
				 0       , 1.0f/S.y, 0       , 0,
				 0       , 0       , 1.0f/S.z, 0,
				 0       , 0       , 0       , 1);
}

/*---------------------------------------------------------------------------*/

mat4f Transformable::getTranslationRotationMatrix()
{
	return mat4f(R.a, R.b, R.c, T.x,
				 R.d, R.e, R.f, T.y,
				 R.g, R.h, R.i, T.z,
				 0	, 0	 , 0  , 1);
}

/*---------------------------------------------------------------------------*/

mat4f Transformable::getLocalTransform()
{
	return mat4f(R.a*S.x, R.b*S.y, R.c*S.z, T.x,
				 R.d*S.x, R.e*S.y, R.f*S.z, T.y,
				 R.g*S.x, R.h*S.y, R.i*S.z, T.z,
				 0		, 0		 , 0	  , 1);
}

/*---------------------------------------------------------------------------*/

mat4f Transformable::getLocalTransformInverse()
{
	vec3f X = vec3f(1.0f / S.x, 1.0f / S.y, 1.0f / S.z);

	return mat4f(R.a*X.x	, R.d*X.y	, R.g*X.z	, -T.x,
				 R.b*X.x	, R.e*X.y	, R.h*X.z	, -T.y,
				 R.c*X.x	, R.f*X.y	, R.i*X.z	, -T.z,
				 0			, 0			, 0			, 1);
}

/*---------------------------------------------------------------------------*/

mat3f Transformable::getNormalMatrix()
{
	vec3f X = vec3f(1.0f / S.x, 1.0f / S.y, 1.0f / S.z);

	return mat3f(R.a*X.x	, R.b*X.x	, R.c*X.x,
				 R.d*X.y	, R.e*X.y	, R.f*X.y,
				 R.g*X.z	, R.h*X.z	, R.i*X.z);
}

/*---------------------------------------------------------------------------*/

mat3f Transformable::getRotation()
{
	return R;
}

/*---------------------------------------------------------------------------*/

vec3f Transformable::getTranslation()
{
	return T;
}

/*---------------------------------------------------------------------------*/

vec3f Transformable::getScale()
{
	return S;
}

/*---------------------------------------------------------------------------*/



