/*
 * Defines a Transformable node.
 * Inherits from Node.
 *
 * A transformation M is composed of rotation, translation and scale. We define
 * M as:
 *
 *    M = T * R * S
 *
 * Transformations are hierachical. This means that every transformable child
 * node will inherit the transformations of its parent. Let's say for instance,
 * that there are two nodes A and B. A is the parent of B and both are
 * transformable. The local transformation of A is denoted MA and for B its MB.
 * The matrix that brings nodes A and B to world space is denoted WA and WB.
 * The resulting transformations for the nodes will then be:
 *
 *    MA = TA * RA * SA
 *    MB = TB * RB * SB
 *
 *    WA = MA = TA * RA * SA
 *    WB = MA * MB = (TA * RA * SA) * (TB * RB * SB)
 *
 * The MB matrix will then, in turn, be recursively be applied to any children
 * of node B to construct new model->world matrices, in the same manner as in
 * the example above.
 *
 */

#ifndef __RC_TRANSFORMABLE_H__
#define __RC_TRANSFORMABLE_H__

class Transformable : public Node {
	friend class SceneGraph;
	friend class World;
	friend class Camera;
	friend class MatrixTransformation;
	friend class TriangleMesh;
	friend class MaterialDirectional;
	friend class MaterialPoint;
	friend class MaterialWeird;

/*----------------------------------------------------------[ Functions ]----*/

	protected:
		Transformable(
				const char	*nname,
				NodeType_t	ntype
			);
		~Transformable();

	public:
		/* Reset the transformation to the identity matrix */
		void reset();

			/* Relative transformations */
		
		void translate(
				f32		x,
				f32		y,
				f32		z
			);

		void translate(
				vec3f	&v
			);

		void scale(
				f32		x,
				f32		y,
				f32		z
			);

		void scale(
				vec3f	&v
			);

		/* Rotate around vector (x, y, z) */

		void rotate(
				f32		angle,
				f32		x,
				f32		y,
				f32		z
			);

		void rotate(
				f32		angle,
				vec3f	&v
			);

			/* Absolute transformations */

		void setTranslate(
				f32		x,
				f32		y,
				f32		z
			);

		void setTranslate(
				vec3f	&v
			);

		void setScale(
				f32		x,
				f32		y,
				f32		z
			);

		void setScale(
				vec3f	&v
			);

		/* Rotate around vector (x, y, z) */

		void setRotate(
				f32		angle,
				f32		x,
				f32		y,
				f32		z
			);

		void setRotate(
				f32		angle,
				vec3f	&v
			);

			/* Useful matrix fetching functions */

		mat3f getRotation();
		vec3f getTranslation();
		vec3f getScale();

		mat4f getTranslationMatrix();
		mat4f getRotationMatrix();
		mat4f getScaleMatrix();

		mat4f getTranslationMatrixInverse();
		mat4f getRotationMatrixInverse();
		mat4f getScaleMatrixInverse();

		mat4f getTranslationRotationMatrix();

		mat4f getLocalTransform();
		mat4f getLocalTransformInverse();

		mat3f getNormalMatrix();

	private:
		void transformCamera(mat4f *matrix);

		void update(
				FrameInfo_t			*di
			);

		void draw(
				FrameInfo_t			*di
			);

/*------------------------------------------------------------[ Members ]----*/

	protected:
		mat3f	R;
		vec3f	T;
		vec3f	S;

		mat4f	world;

};


#endif /* __RC_TRANSFORMABLE_H__ */

