#include "RenderChimp.h"
#include "SceneGraph.h"
#include "Geometry.h"
#include "TriangleMesh.h"
#include "PwnParser.h"

static const Result_t WRN_LGT_NO_PARENT = {
	RESULT_WARNING, "The light node has no parent" };

static u16		trimesh_id = 0;

/*---------------------------------------------------------------------------*/

TriangleMesh::TriangleMesh(
		char	*nname
	) : Geometry(nname, NODE_TRIANGLEMESH)
{
	vertexArray = 0;
	indexArray = 0;
	material = 0;
}

/*---------------------------------------------------------------------------*/

TriangleMesh::TriangleMesh(
		char					*nname,
		char					*filename,

		bool					reload_resources,
		u32						purge_level
	) : Geometry(nname, NODE_TRIANGLEMESH)
{
	TriangleMesh *m;
	Group *g = sceneGraph.createGroup("temp_group");

	PwnParse(g, filename, true, reload_resources, purge_level);

	m = (TriangleMesh *) g->getChild(0);

	this->vertexArray = (VertexArray *) m->getVertexArray()->instantiate();
	this->indexArray = (IndexArray *) m->getIndexArray()->instantiate();
	this->material = (Material *) m->getMaterial()->instantiate();

	sceneGraph.deleteNode(g);
}

/*---------------------------------------------------------------------------*/

TriangleMesh::TriangleMesh(
		char					*nname,

		VertexArray				*vertex_array,
		IndexArray				*index_array,
		Material				*nmaterial,

		const bool				instantiate_resources
	) : Geometry(nname, NODE_TRIANGLEMESH)
{
	if (instantiate_resources) {
		vertexArray = (VertexArray *) vertex_array->instantiate();
		indexArray = (IndexArray *) index_array->instantiate();
		material = (Material *) nmaterial->instantiate();
	} else {
		vertexArray = vertex_array;
		indexArray = index_array;
		material = nmaterial;
	}
}

/*---------------------------------------------------------------------------*/

TriangleMesh::TriangleMesh(
		char					*nname,

		vec3f					*position_array,
		vec2f					*texcoord_array,
		vec3f					*normal_array,
		vec4f					*color_array,
		const u32				n_vertices,

		u16						*index_array,
		const u32				n_indices,
		IndexArrayLayout_t		index_layout,

		u32						purge_level
	) : Geometry(nname, NODE_TRIANGLEMESH)
{
	vertexArray = 0;
	indexArray = 0;
	material = 0;

	setVertexArray(position_array,
		texcoord_array,
		normal_array,
		color_array,
		n_vertices,
		purge_level);

	setIndexArray(index_array,
		n_indices,
		index_layout,
		purge_level);
}

/*---------------------------------------------------------------------------*/

TriangleMesh::~TriangleMesh()
{

	ensure_heap();

	if (vertexArray)
		sceneGraph.dereferenceResource(vertexArray);
	
	ensure_heap();

	if (indexArray)
		sceneGraph.dereferenceResource(indexArray);

	ensure_heap();

	if (material)
		sceneGraph.dereferenceResource(material);

	ensure_heap();

}

/*---------------------------------------------------------------------------*/

void TriangleMesh::drawSelf(
		FrameInfo_t			*di
	)
{
	if (!material) {
		return;
	}

	GL_ASSERT();

	/* Sets active shader program */
	material->prepareDrawing(di, this);

	prepareDrawing(di);

	di->matTransform->setTransformationMatrices(material->shaderProgram);

	vertexArray->pushToShader(material->shaderProgram);

	if (indexArray) {
		glDrawElements(indexArray->getGLPrimitiveLayout(), indexArray->getLength(),
			GL_UNSIGNED_SHORT, indexArray->getArray());

	} else {
		glDrawArrays(GL_TRIANGLES, 0, vertexArray->getLength());

	}

	GL_ASSERT();
}

/*---------------------------------------------------------------------------*/

void TriangleMesh::setVertexArray(
		VertexArray				*vertex_array,

		bool					instantiate_resource
	)
{
	VertexArray::setVertexArray(&vertexArray, vertex_array, instantiate_resource);
}

/*---------------------------------------------------------------------------*/

void TriangleMesh::setVertexArray(
		vec3f					*position_array,
		vec2f					*texcoord_array,
		vec3f					*normal_array,
		vec4f					*color_array,
		u32						n_vertices,

		u32						purge_level
	)
{

	u32 i;

	/* Element lengths */
	u32 pl = 0, tl = 0, nl = 0, cl = 0;

	/* Element offsets */
	u32 ps = 0, ts = 0, ns = 0, cs = 0;

	u32 stride;

	f32	*va; /* Temp array */

	/* Set up new identifier */
	char id[9] = "tm";
	sprintf(&id[2], "%4xv\0", trimesh_id++);

	if (vertexArray)
		sceneGraph.dereferenceResource(vertexArray);

	/* Determine stride and offset for different attributes */
	if (position_array)
		pl = 3;

	if (texcoord_array)
		tl = 2;

	if (normal_array)
		nl = 3;

	if (color_array)
		cl = 4;

	stride = pl + tl + nl + cl;

	/* Sum offsets */
	ps = 0;
	ts = ps + pl;
	ns = ts + tl;
	cs = ns + nl;

	/* Populate array */

	va = (f32 *) malloc(sizeof(f32) * n_vertices * stride);

	if (position_array) {
		u32 j = ps;
		for (i = 0; i < n_vertices; i++) {
			va[j    ] = position_array[i].x;
			va[j + 1] = position_array[i].y;
			va[j + 2] = position_array[i].z;
			j += stride;
		}
	}

	if (texcoord_array) {
		u32 j = ts;
		for (i = 0; i < n_vertices; i++) {
			va[j    ] = texcoord_array[i].u;
			va[j + 1] = texcoord_array[i].v;
			j += stride;
		}
	}

	if (normal_array) {
		u32 j = ns;
		for (i = 0; i < n_vertices; i++) {
			va[j    ] = normal_array[i].x;
			va[j + 1] = normal_array[i].y;
			va[j + 2] = normal_array[i].z;
			j += stride;
		}
	}

	if (color_array) {
		u32 j = cs;
		for (i = 0; i < n_vertices; i++) {
			va[j    ] = color_array[i].r;
			va[j + 1] = color_array[i].g;
			va[j + 2] = color_array[i].b;
			va[j + 3] = color_array[i].a;
			j += stride;
		}
	}

	/* Create the vertex array resource */
	vertexArray = sceneGraph.createVertexArray(id, va, stride,
		n_vertices, purge_level);

	/* Set attribute names and offsets */
	if (position_array)
		vertexArray->setAttribute(ps, pl, "Vertex");

	if (texcoord_array)
		vertexArray->setAttribute(ts, tl, "TexCoord");

	if (color_array)
		vertexArray->setAttribute(cs, cl, "Color");

	if (normal_array)
		vertexArray->setAttribute(ns, nl, "Normal");

}

/*---------------------------------------------------------------------------*/

void TriangleMesh::setIndexArray(
		IndexArray				*index_array,

		bool					instantiate_resource
	)
{
	IndexArray::setIndexArray(&indexArray, index_array, instantiate_resource);
}

/*---------------------------------------------------------------------------*/

void TriangleMesh::setIndexArray(
		u16						*index_array,
		u32						n_indices,

		IndexArrayLayout_t		layout,

		u32						purge_level
	)
{
	/* Set up new identifier */
	char id[9] = "tm";
	sprintf(&id[2], "%xv\0", trimesh_id++);

	if (indexArray)
		sceneGraph.dereferenceResource(indexArray);

	indexArray = sceneGraph.createIndexArray(id, index_array,
		n_indices, layout, purge_level);

}

/*---------------------------------------------------------------------------*/

void TriangleMesh::prepareDrawing(
		FrameInfo_t				*di
	)
{
	GLint loc;

	ShaderProgram *shaderProgram = material->shaderProgram;

	loc = shaderProgram->getUniformLocation("NormalMatrix");

	if (loc >= 0) {
		di->matTransform->normal = getNormalMatrix();
		glUniformMatrix3fv(loc, 1, GL_FALSE, di->matTransform->normal.s);
	}

	loc = shaderProgram->getUniformLocation("ViewPosition");

	if (loc >= 0)
		glUniform3fv(loc, 1, di->matTransform->viewPosition.s);

	GL_ASSERT();
}

/*---------------------------------------------------------------------------*/

void TriangleMesh::setMaterial(
		Material				*nmaterial,

		bool					instantiate_resource
	)
{
	GL_ASSERT();

	if (material)
		sceneGraph.dereferenceResource(material);

	GL_ASSERT();

	if (instantiate_resource) {
		material = (Material *) nmaterial->instantiate();
	} else {
		material = nmaterial;
	}

	GL_ASSERT();
}

/*---------------------------------------------------------------------------*/

Material *TriangleMesh::getMaterial()
{
	return material;
}

/*---------------------------------------------------------------------------*/

VertexArray *TriangleMesh::getVertexArray()
{
	return vertexArray;
}

/*---------------------------------------------------------------------------*/

IndexArray *TriangleMesh::getIndexArray()
{
	return indexArray;
}

/*---------------------------------------------------------------------------*/

Node *TriangleMesh::duplicateSelf()
{
	u32 i;

	TriangleMesh *c = sceneGraph.createTriangleMesh(name,
		vertexArray, indexArray, material, true);

	for (i = 0; i < MAX_SHADER_LIGHT_SLOTS; i++) {
		if (light[i] != 0) {
			c->setLight(light[i], i);
		}
	}

	return c;
}

/*---------------------------------------------------------------------------*/

