/*
 * Vertex array.
 * Inherits from Resource.
 *
 * This is an array of vertices that is used in conjunction with an index
 * array. The index array points to vertex attributes from this array, and
 * these will be used to construct triangles with accompanying data.
 *
 * A vertex array should at least contain an attribute that specifies the
 * position of a vertex. Other common properties are normals, color and texture
 * coordinates.
 *
 * The vertex array is organized in a array of structs-manner, as opposed to a
 * struct of arrays. The difference is illustrated below with a vertex array
 * containing position, normal and color data.
 *
 *                       Vertex attributes
 *
 *   Struct of arrays (not used)       Array of structs
 *   (SOA)                             (AOS)
 *
 *    Vertex[0] Position.x              Vertex[0] Position.x       ^
 *    Vertex[0] Position.y              Vertex[0] Position.y       | S
 *    Vertex[0] Position.z              Vertex[0] Position.z       | t
 *    Vertex[1] Position.x              Vertex[0] Normal.x         | r
 *    Vertex[1] Position.y              Vertex[0] Normal.y         | i
 *    Vertex[1] Position.z              Vertex[0] Normal.z         | d
 *     ...                              Vertex[0] Color.r          | e
 *    Vertex[0] Normal.x                Vertex[0] Color.g          |
 *    Vertex[0] Normal.y                Vertex[0] Color.b          |
 *    Vertex[0] Normal.z                Vertex[0] Color.a          v
 *    Vertex[1] Normal.x                Vertex[1] Position.x
 *    Vertex[1] Normal.y				Vertex[1] Position.y
 *    Vertex[1] Normal.z				Vertex[1] Position.z
 *     ...								Vertex[1] Normal.x
 *     ...								Vertex[1] Normal.y
 *    Vertex[0] Color.r					Vertex[1] Normal.z
 *    Vertex[0] Color.g					Vertex[1] Color.r
 *    Vertex[0] Color.b					Vertex[1] Color.g
 *    Vertex[0] Color.a					Vertex[1] Color.b
 *    Vertex[1] Color.r					Vertex[1] Color.a
 *    Vertex[1] Color.g                  ...
 *    Vertex[1] Color.b
 *    Vertex[1] Color.a
 *     ...
 *
 * In this example, for the AOS the stride = 10, since the data type repeats
 * after every 10 entries.
 *
 * The offset for the "Normal" attribute in this example is 3, since it's three
 * entries in from the beginning of the array. It's length is 3, since it has
 * three components (x, y, z).
 */

#ifndef __RC_VERTEXARRAY_H__
#define __RC_VERTEXARRAY_H__

static u16 va_clone_counter		= 0;


class VertexArray : public Resource {
	friend class SceneGraph;
	friend class TriangleMesh;
	friend class Sprite;

/*----------------------------------------------------------[ Functions ]----*/

	private:
		VertexArray(
				const char				*nidentifier,

				f32						*new_array,
				u32						nstride,		/* Length of each
														 * struct.
														 */
				u32						n,				/* Number of structs */

				u32						purge_level
			);

		virtual ~VertexArray();

	public:
		void setAttribute(
				u32						offset,	/* Offset to first entry in
												 * the array.
												 */
				u32						length, /* Number of components */
				const char				*name	/* Name of the attribute. This
												 * must comply with the vertex
												 * attribute names in the
												 * shader this vertex array is
												 * intended to be used with.
												 */
			);

		Resource *clone(
				const char				*clone_name,

				u32						purge_level
			);

		f32	*getArray();

		u32 getLength();

		u32 getNAttributes();

		char *getAttributeName(
				u32						i
			);

		/* If there are no normals in the vertex array, they can be generated
		 * by calling this function. For a more exact result, however, it is
		 * better to provide normals with the mesh, if possible.
		 *
		 * Normals are generated with regard to their adjacent triangles, so an
		 * index array is needed to specify which those triangles are.
		 */
		void generateNormals(
				IndexArray				*indexArray
			);

	private:

		/* Static helper functions */
		static void setVertexArray(
				VertexArray				**target_vertex_array,
				VertexArray				*new_vertex_array,

				bool					instantiate_resource
			);

		/* Prepare this index array for rendering */
		void pushToShader(
				ShaderProgram			*shader
			);

#ifdef DEBUG
	protected:
		void print(
				FILE	*f
			);
#endif /* DEBUG */

/*------------------------------------------------------------[ Members ]----*/

	private:
		f32		*vertexArray;
		u32		n_elements;
		u32		stride;

		/* Attributes */
		u32		n_attribs;
		u32		attrib_offset[MAX_ATTRIBS];
		u32		attrib_length[MAX_ATTRIBS];
		char	*attrib_name[MAX_ATTRIBS];

};

#endif /* __RC_VERTEXARRAY_H__ */

