
#ifdef WIN32

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

#include <SDL/SDL.h>
#include <SDL/SDL_syswm.h>
#include <SDL/SDL_image.h>

#include <EGL/egl.h>
#include <GLES2/gl2.h>

#include "RenderChimp.h"
#include "SceneGraph.h"

#include "ape.h"

#ifdef _DEBUG 
	#define CRTDBG_MAP_ALLOC 
	#include <crtdbg.h> 
#endif

#define SPLASH_TIME				7.0f


static EGLDisplay eglDisplay;
static EGLConfig eglConfig;
static EGLSurface eglSurface;
static EGLContext eglContext;

static SDL_Surface *screen;
static SDL_Surface *skin;

static GLuint skin_tex;

static f64 timer_freq;

void (*init_func)();
unsigned int (*update_func)();
void (*destroy_func)();

static int esc_pressed = 0;

static bool portrait_orientation = false;

#define MEM_REFRESH			0.10f
#define MEM_STEPS			512

static f32 mem_size_visualizer_vertices[MEM_STEPS * 4];
static u32 max_mem = 1;

static u32 mem_list[MEM_STEPS];

#define VIEW_WIDTH			320
#define VIEW_HEIGHT			480

#define VIEW_OFFSET_P_X		41
#define VIEW_OFFSET_P_Y		119

#define VIEW_OFFSET_L_X		142
#define VIEW_OFFSET_L_Y		41


static i32 skinvp_w, skinvp_h;
static i32 skinvp_x, skinvp_y;
static i32 vp_w, vp_h;
static i32 vp_x, vp_y;
static i32 slack_w, slack_h;
static i32 slack_x0, slack_y0;
static i32 slack_x1, slack_y1;

static const char *mem_vshader = 
	"attribute vec2 pos;\n"
	"varying mediump vec2 p;\n"
	"uniform int is_portrait;\n"
	"void main() {\n"
	"	if (is_portrait != 0) {\n"
	"		gl_Position = vec4(pos.y, -pos.x, 1.0, 1.0);\n"
	"	} else {\n"
	"		gl_Position = vec4(pos.x, pos.y, 1.0, 1.0);\n"
	"	}\n"
	"	p = (pos + 1.0) * 0.5;\n"
	"}\n";

static const char *mem_fshader = 
	"varying mediump vec2 p;\n"
	"uniform int is_portrait;\n"
	"void main() {\n"
	"	gl_FragColor = vec4(0.0, 0.0, 1.0 - p.x * 0.7, 1.0);\n"
	"}\n";

static const char *clear_vshader = 
	"attribute vec2 pos;\n"
	"void main() {\n"
	"	gl_Position = vec4(pos.x, pos.y, 1.0, 1.0);\n"
	"}\n";

static const char *clear_fshader = 
	"void main() {\n"
	"	gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);\n"
	"}\n";



static const char *vshader = 
	"attribute vec2 pos;\n"
	"attribute vec2 tex;\n"
	"varying mediump vec2 tt;\n"
	"void main() {\n"
	"	gl_Position = vec4(pos.x, pos.y, 1.0, 1.0);\n"
	"	tt = tex;\n"
	"}\n";

static const char *fshader = 
	"varying mediump vec2 tt;\n"
	"uniform sampler2D smp2d;\n"
	"void main() {\n"
	"	gl_FragColor = texture2D(smp2d, tt);\n"
	"}\n";

GLfloat screen_quad_pos[8] = {
	 1.0f, -1.0f,
	1.0f, 1.0f,
	-1.0f, -1.0f,
	-1.0f, 1.0f};

GLfloat screen_quad_tex[8] = {
	1.0f, 1.0f,
	1.0f, 0.0f,
	0.0f, 1.0f,
	0.0f, 0.0f};

GLfloat screen_quad_rot_tex[8] = {
	0.0f, 1.0f,
	1.0f, 1.0f,
	0.0f, 0.0f,
	1.0f, 0.0f};

GLfloat *screen_tex = 0;




static void toggle_orientation()
{
	portrait_orientation = !portrait_orientation;
	i32 res_x, res_y;
	f32 *p;

	if (portrait_orientation) {
		res_x = VIEW_WIDTH;
		res_y = VIEW_HEIGHT;
		vp_x = VIEW_OFFSET_P_X;
		vp_y = VIEW_OFFSET_P_Y;
		skinvp_w = skin->w;
		skinvp_h = skin->h;
		skinvp_x = skinvp_h / 2 - skinvp_w / 2;
		skinvp_y = 0;

		slack_w = skinvp_x;
		slack_h = skinvp_h;
		slack_x0 = 0;
		slack_y0 = 0;
		slack_x1 = skinvp_x + skinvp_w;
		slack_y1 = 0;

		screen_tex = screen_quad_tex;
	} else {
		res_x = VIEW_HEIGHT;
		res_y = VIEW_WIDTH;
		vp_x = VIEW_OFFSET_L_X;
		vp_y = VIEW_OFFSET_L_Y;
		skinvp_w = skin->h;
		skinvp_h = skin->w;
		skinvp_x = 0;
		skinvp_y = skinvp_w / 2 - skinvp_h / 2;

		slack_w = skinvp_w;
		slack_h = skinvp_y;
		slack_x0 = 0;
		slack_y0 = skinvp_y + skinvp_h;
		slack_x1 = 0;
		slack_y1 = 0;

		screen_tex = screen_quad_rot_tex;
	}

	sceneGraph.resize(res_x, res_y);

	vp_w = res_x;
	vp_h = res_y;
	
	vp_x += skinvp_x;
	vp_y += skinvp_y;

	p = mem_size_visualizer_vertices;
/*
	for (i = 0; i < MEM_STEPS; i++) {
		f32 t;

		t = p[0];
		p[0] = p[1];
		p[1] = t;

		t = p[2];
		p[2] = p[3];
		p[3] = t;

		p += 4;
	}
*/
}


static void init_sdl(int w, int h)
{
	int result;

	atexit(SDL_Quit);

	result = SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER);
	if (result < 0)
		return;

	screen = SDL_SetVideoMode(h, h, 24, SDL_HWSURFACE | SDL_DOUBLEBUF);
	if (!screen)
		return;

}

static void let_there_be_light()
{
	int iConfigs;
	EGLint iMajorVersion, iMinorVersion;

	HWND hWnd = 0;
	HDC hDC = 0;
	u64 tf;

	NativeWindowType eglWindow = 0;

	SDL_SysWMinfo wmInfo;
	

	/* Set up timer */
	HANDLE process_handle = GetCurrentProcess();
	SetProcessAffinityMask(process_handle, 1);

	QueryPerformanceFrequency((LARGE_INTEGER *) &tf);
	timer_freq = (f64) tf;

	/* Init OpenGL ES 2.0 */
	eglDisplay = 0;
	eglConfig = 0;
	eglSurface = 0;
	eglContext = 0;

	skin = IMG_Load("../skin/iphone.png");


	init_sdl(skin->w, skin->h);

		/* Get window context */
	SDL_VERSION(&wmInfo.version);
	SDL_GetWMInfo(&wmInfo);

	hWnd = (NativeWindowType) wmInfo.window;
	hDC = GetDC(hWnd);

	eglWindow = hWnd;

	hDC = GetDC(hWnd);
	if (!hDC)
		return;


	eglDisplay = eglGetDisplay((NativeDisplayType) hDC);

	if (eglDisplay == EGL_NO_DISPLAY)
         eglDisplay = eglGetDisplay((NativeDisplayType) EGL_DEFAULT_DISPLAY);

	if (!eglInitialize(eglDisplay, &iMajorVersion, &iMinorVersion))
		return;

	const EGLint pi32ConfigAttribs[] =
	{
		EGL_LEVEL,				0,
		EGL_SURFACE_TYPE,		EGL_WINDOW_BIT,
		EGL_RENDERABLE_TYPE,	EGL_OPENGL_ES2_BIT,
		EGL_NATIVE_RENDERABLE,	EGL_FALSE,
		EGL_DEPTH_SIZE,			16,
		EGL_NONE
	};

	if (!eglChooseConfig(eglDisplay, pi32ConfigAttribs, &eglConfig, 1, &iConfigs))
		return;

	if (iConfigs != 1)
		return;

	eglSurface = eglCreateWindowSurface(eglDisplay, eglConfig, eglWindow, NULL);

	if (eglSurface == EGL_NO_SURFACE) {
		eglGetError(); // Clear error
		eglSurface = eglCreateWindowSurface(eglDisplay, eglConfig, NULL, NULL);
	}

	eglContext = eglCreateContext(eglDisplay, eglConfig, NULL, NULL);
	eglMakeCurrent(eglDisplay, eglSurface, eglSurface, eglContext);

	glViewport(0, 0, skin->w, skin->h);


}

/* Timer functions */
static void get_time(
		DeviceState_t	*ds
	)
{
	u64 t;
	QueryPerformanceCounter((LARGE_INTEGER *) &t);
	f32 newTime = (f32) ((f64) t / timer_freq);
	ds->timeStep = newTime - ds->time;
	ds->time = newTime;
}

u32 key_pressed = 0;
u32 mouse_pressed = 0;

static int get_input(
		DeviceState_t	*ds
	)
{
	SDL_Event event;
	i32 more;

	more = SDL_PollEvent(&event);

	while (more) {
		if (event.key.type == SDL_QUIT) {
			return 1;
		} else if (event.key.type == SDL_KEYDOWN && !key_pressed) {
			key_pressed = 1;

			switch (event.key.keysym.sym) {
				case SDLK_TAB:
					toggle_orientation();
					break;
				case SDLK_ESCAPE:
					esc_pressed = 1;
					break;
				default:
					break;
			}

		} else if (event.key.type == SDL_KEYUP) {
			key_pressed = 0;
		} else if (event.key.type == SDL_MOUSEBUTTONDOWN || (event.key.type == SDL_MOUSEMOTION && mouse_pressed)) {
			if (event.motion.x >= vp_x && event.motion.x <= vp_x + vp_w &&
				(skin->h - event.motion.y) >= vp_y && (skin->h - event.motion.y) <= vp_y + vp_h) {

				ds->x = 2.0f * ((((f32) event.motion.x) - vp_x) / vp_w - 0.5f);
				ds->y = 2.0f * ((skin->h - ((f32) event.motion.y) - vp_y) / vp_h - 0.5f);

				mouse_pressed = 1;
			}
		} else if (event.key.type == SDL_MOUSEBUTTONUP) {
			mouse_pressed = 0;
		}

		more = SDL_PollEvent(&event);
	}

	ds->touch = mouse_pressed ? APE_TOUCH_SINGLE : APE_TOUCH_NONE;

	return 0;
}


static void update_mem_info()
{
	f32 *p = mem_size_visualizer_vertices;
	u32 i;
	u32 offset = 0;
	max_mem = 0;

	for (i = 0; i < MEM_STEPS - 1; i++) {
		mem_list[i] = mem_list[i + 1];

		if (mem_list[i] > max_mem)
			max_mem = mem_list[i];

	}

	mem_list[MEM_STEPS - 1] = getAllocationPeakLocal();

	if (mem_list[MEM_STEPS - 1] > max_mem) {
		max_mem = mem_list[MEM_STEPS - 1];
	}

	p = mem_size_visualizer_vertices;

	for (i = 0; i < MEM_STEPS; i++) {
		p[0 + offset] = 1.0f - (2.0f * ((f32) mem_list[i]) / ((f32) max_mem));
		p += 4;
	}
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	unsigned int quit = 0;
	GLuint vsid, fsid, pid, mem_pid, c_pid;
	GLint loc;
	u32 i;
	f32 *p;
	f32 next_mem_refresh = 0.0f;
	char mem_size_buf[64];

	DeviceState_t ds;
	memset(&ds, 0, sizeof(DeviceState_t));

	p = mem_size_visualizer_vertices;

	for (i = 0; i < MEM_STEPS; i++) {
		f32 f = 2.0f * ((f32) i) / ((f32) MEM_STEPS) - 1.0f;
		p[0] = 1.0f;
		p[1] = f;
		p[2] = 1.0f;
		p[3] = f;

		p += 4;
	}

	memset(mem_list, 0, sizeof(u32) * MEM_STEPS);

	ResultInit();

	let_there_be_light();

	sceneGraph.init();


	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

	vsid = glCreateShader(GL_VERTEX_SHADER);
	fsid = glCreateShader(GL_FRAGMENT_SHADER);

	GL_ASSERT();

	glShaderSource(vsid, 1, (const char **) &vshader, NULL);
	glShaderSource(fsid, 1, (const char **) &fshader, NULL);

	GL_ASSERT();

	glCompileShader(vsid);
	glCompileShader(fsid);

	GL_ASSERT();

	pid = glCreateProgram();
	GL_ASSERT();

	glAttachShader(pid, vsid);
	glAttachShader(pid, fsid);
	GL_ASSERT();

	GL_ASSERT();

	glLinkProgram(pid);
	GL_ASSERT();

	glUseProgram(pid);




	vsid = glCreateShader(GL_VERTEX_SHADER);
	fsid = glCreateShader(GL_FRAGMENT_SHADER);

	GL_ASSERT();

	glShaderSource(vsid, 1, (const char **) &mem_vshader, NULL);
	glShaderSource(fsid, 1, (const char **) &mem_fshader, NULL);

	GL_ASSERT();

	glCompileShader(vsid);
	glCompileShader(fsid);

	GL_ASSERT();

	mem_pid = glCreateProgram();
	GL_ASSERT();

	glAttachShader(mem_pid, vsid);
	glAttachShader(mem_pid, fsid);
	GL_ASSERT();

	GL_ASSERT();

	glLinkProgram(mem_pid);




	vsid = glCreateShader(GL_VERTEX_SHADER);
	fsid = glCreateShader(GL_FRAGMENT_SHADER);

	GL_ASSERT();

	glShaderSource(vsid, 1, (const char **) &clear_vshader, NULL);
	glShaderSource(fsid, 1, (const char **) &clear_fshader, NULL);

	GL_ASSERT();

	glCompileShader(vsid);
	glCompileShader(fsid);

	GL_ASSERT();

	c_pid = glCreateProgram();
	GL_ASSERT();

	glAttachShader(c_pid, vsid);
	glAttachShader(c_pid, fsid);
	GL_ASSERT();

	GL_ASSERT();

	glLinkProgram(c_pid);






	glDisable(GL_DEPTH_TEST);
	GL_ASSERT();

	glGenTextures(1, &skin_tex);
	glBindTexture(GL_TEXTURE_2D, skin_tex);

	GL_ASSERT();

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, skin->w, skin->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, skin->pixels);

	GL_ASSERT();

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	GL_ASSERT();


	/*
	loc = glGetAttribLocation(pid, "tex");
	GL_ASSERT();
	glVertexAttribPointer(loc, 2, GL_FLOAT, GL_FALSE, 0, screen_quad_tex);
	GL_ASSERT();
	glEnableVertexAttribArray(loc);

	GL_ASSERT();

	loc = glGetAttribLocation(pid, "pos");
	GL_ASSERT();
	glVertexAttribPointer(loc, 2, GL_FLOAT, GL_FALSE, 0, screen_quad_pos);
	GL_ASSERT();
	glEnableVertexAttribArray(loc);

	GL_ASSERT();
	
	loc = glGetUniformLocation(pid, "smp2d");

	GL_ASSERT();

	glUniform1i(loc, 0);
*/

	{
		static char *welcome = {
			"~c5369~bRenderChimp~b~c9999 (v ~i0.2a~i)\n\n"\
			"  Written by ~sMr. Andersson~s (magnusa@cs.lth.se)\n\n\n\n"\
			"Press '~c2929tab~c9999' to swap between portrait and landscape orientation.\n\n"\
			"Press '~c2929esc~c9999' to exit.\n\n\n\n\n"\
			"Press any key to continue..."
		};
		f32 splash_time, t;

		get_time(&ds);
		ds.timeStep = 0.0f;

		splash_time = ds.time + SPLASH_TIME;
		t = 0.0f;

		glViewport(0, 0, skin->h, skin->h);
		
		sceneGraph.resize(skin->h, skin->h);

		while (t < splash_time) {
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

			sceneGraph.text->print(welcome, -0.8f, 0.8f, 450, 20);

			sceneGraph.text->printAll();

			eglSwapBuffers(eglDisplay, eglSurface);

			if (key_pressed || mouse_pressed)
				break;

			get_input(&ds);

			get_time(&ds);
			t = ds.time;
		}

	}

	toggle_orientation();

	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);

	RCInit();

	for (;;) {
		sceneGraph.resize(vp_w, vp_h);

		/* Draw app */
		glViewport(vp_x, vp_y, vp_w, vp_h);

		if (get_input(&ds))
			break;

		get_time(&ds);

		quit = RCUpdate(&ds);

		if (quit | esc_pressed)
			break;

		sceneGraph.text->printAll();

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable(GL_CULL_FACE);
		glDisable(GL_DEPTH_TEST);
		glActiveTexture(GL_TEXTURE0);

		GL_ASSERT();

		/* Draw skin */
		glViewport(skinvp_x, skinvp_y, skinvp_w, skinvp_h);
		
		glUseProgram(pid);
	
		glBindTexture(GL_TEXTURE_2D, skin_tex);
		
		loc = glGetUniformLocation(pid, "smp2d");
		glUniform1i(loc, 0);

		loc = glGetAttribLocation(pid, "tex");
		glVertexAttribPointer(loc, 2, GL_FLOAT, GL_FALSE, 0, screen_tex);
		glEnableVertexAttribArray(loc);

		loc = glGetAttribLocation(pid, "pos");
		glVertexAttribPointer(loc, 2, GL_FLOAT, GL_FALSE, 0, screen_quad_pos);
		glEnableVertexAttribArray(loc);

		glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

		if (next_mem_refresh < ds.time) {
			update_mem_info();
			next_mem_refresh = ds.time + MEM_REFRESH;
		}

		sceneGraph.text->printAll();

		/* Draw mem info */

		glUseProgram(c_pid);

		loc = glGetAttribLocation(c_pid, "pos");
		glVertexAttribPointer(loc, 2, GL_FLOAT, GL_FALSE, 0, screen_quad_pos);
		glEnableVertexAttribArray(loc);


		glViewport(slack_x1, slack_y1, slack_w, slack_h);
		glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

		glViewport(slack_x0, slack_y0, slack_w, slack_h);
		glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

		glUseProgram(mem_pid);

		loc = glGetAttribLocation(mem_pid, "pos");
		glVertexAttribPointer(loc, 2, GL_FLOAT, GL_FALSE, 0, mem_size_visualizer_vertices);
		glEnableVertexAttribArray(loc);

		loc = glGetUniformLocation(mem_pid, "is_portrait");
		glUniform1i(loc, portrait_orientation ? 0 : 1);

		glDrawArrays(GL_TRIANGLE_STRIP, 0, MEM_STEPS * 2);

		sprintf(mem_size_buf, "Allocated:\n %d B\n\nPeak:\n %d B\n", getAllocationPeakLocal(), getAllocationPeak());
		
		GL_ASSERT();

		sceneGraph.resize(slack_w, slack_h);
		sceneGraph.text->print(mem_size_buf, -0.9f, 0.5f, 250, 5);

		sceneGraph.text->printAll();

		/* Draw errors and warnings */
		glViewport(0, 0, skin->h, skin->h);
		sceneGraph.resize(skin->h, skin->h);
		ResultUpdateMessage(skin->h - 50);

		sceneGraph.text->printAll();

		eglSwapBuffers(eglDisplay, eglSurface);

		GL_ASSERT();
	}

	RCDestroy();

	sceneGraph.destroy();

	ResultRelease();

	if (getAllocationSize() != 0 || getAllocationCount() != 0) {
		logAllocations();
	}

	#ifdef _DEBUG
		_CrtSetDbgFlag(_CRTDBG_LEAK_CHECK_DF);
	#endif

	return 0;
}


/* Texture functions */

void apeGetTexture(
		const char			*filename,
		u8					**pixels,
		u32					*width,
		u32					*height,
		u32					*bytes_per_pixel
	)
{
	u32 s;
	SDL_Surface *img = IMG_Load(filename);

	if (!img) {
		*pixels = 0;
		return;
	}

	*width = img->w;
	*height = img->h;
	*bytes_per_pixel = img->format->BytesPerPixel;

	s = (*bytes_per_pixel) * (*width) * (*height);

	*pixels = (u8 *) malloc(s);
	memcpy(*pixels, img->pixels, s);

	SDL_FreeSurface((SDL_Surface *) img);
}

void apeDestroyTexture(
		unsigned char		*pixels
	)
{
	free(pixels);
}


/*
void apeGetAccelerometer(
		f32					*roll,
		f32					*pitch,
		f32					*facing
	)
{
	*roll = 0.0f;
	*pitch = 0.0f;
	*facing = -0.5f;
}

*/


void apeSetViewport(
		i32					x,
		i32					y,
		u32					width,
		u32					height
	)
{
	glViewport(vp_x + x, vp_y + y, width, height);
}

void apeGetDisplaySize(
		u32					*width,
		u32					*height
	)
{
	*width = vp_w;
	*height = vp_h;
}

#endif /* WIN32 */


